/*
  -------------------------------
  laser_anemometer_sender_v5.ino
  -------------------------------
  Author: lingib
  https://www.instructables.com/member/lingib/instructables/
  License:  GPL-3.0
  Date:     22 November 2021

  About:
   - This program uses two VL6180X Time Of light (TOF) sensors to measure how far off-centre a pipe is from vertical.
   - Wind speed and Direction are calculated from the off-centre coordinates
   - These measurements are then sent to Processing3 "laser_anemometer_display.pde" for display

  Caveat:
   - The wind speed and direction readings are only as accurate as the sensors.
   - The VL6180X sensor resolution is +/- 1mm which causes th ereadings to fluctuate slightly in still air.

  Code assumes:
   - A 5V arduino UNO R3
   - The VL6810X distance modules have onboard I2C voltage convertors and a level-shifted SHDN pin

  Wiring:
   - Arduino pin A2 is attached to The North-facing SHDN (shutdown) pin
   - Arduino pin A3 is attached to The East-facing  SHDN (shutdown) pin
   - Arduino pin A4 is attached to the SDA pin of both VL6810X modules
   - Arduino pin A5 is attached to the SCL pin of both VL6810X modules
*/

// ----- libraries
#include <Wire.h>
#include <VL6180X.h>

// ------ Create a new instance for each sensor
VL6180X North;                                        // North-facing sensor
VL6180X East;                                         // East-facing sensor

// ----- sensor override addresses
/*
  Multiple sensors can be used by turning each sensor ON
  in turn and overiding the default address of 0x29
*/
const byte North_address = 0x30;
const byte East_address = 0x31;

// ----- These Arduino pins must be wired to the VL6180X SHDN pin
const int North_enable = 16;                          // arduino pin A2
const int East_enable = 17;                           // arduino pin A3

// ----- sensor info
const float Radius = 36;                              // inside pipe radius (mm)
const float North_offset = -13.0;                     // distance (mm) from centre-line
const float East_offset = -13.0 ;                     // distance (mm) from centre-line
const float Scale = 4.0;                              // scale factor
float North_reading = 0.0;
float East_reading = 0.0;
float Windspeed = 0.0;
float Direction = 0.0;

// ----- Filter0 (North Sensor)
float Array0[40] = {0};                               // array
int Index0 = 0;                                       // position of first item
float Sum0 = 0.0;                                     // array sum
float Average0 = 0.0;                                 // moving average

// ----- Filter1 (East Sensor)
float Array1[40] = {0};                               // array
int Index1 = 0;                                       // position of first item
float Sum1 = 0.0;                                     // array sum
float Average1 = 0.0;                                 // moving average

// ----- Filter2 (Direction)
float Array2[40] = {0};                               // array
int Index2 = 0;                                       // position of first item
float Sum2 = 0.0;                                     // array sum
float Average2 = 0.0;                                 // moving average

// ----- Communications
const long Baudrate = 115200;                         // serial comminication speed
char Char;

// ---------------
//  setup()
// ---------------
void setup() {
  // ----- Initialise communications
  Serial.begin(115200);
  Wire.begin();

  // ----- Configure Arduino pins
  pinMode(North_enable, OUTPUT);
  pinMode(East_enable, OUTPUT);

  // ----- Shutdown sensors
  digitalWrite(North_enable, LOW);
  digitalWrite(East_enable, LOW);

  // ----- Configure North_facing sensor
  digitalWrite(North_enable, HIGH);                    // activate sensor
  delay(50);
  North.init();                                       // configure sensor
  North.configureDefault();
  North.setAddress(North_address);                    // overide the default address
  North.setTimeout(500);
  /*
    ------------
    Calibration:
    ------------
    The VL6180X readings can be up to 15mm out
    To calibrate the sensors:
     - Comment out code lines 153, 162..168, 198..199
     - Uncomment/comment the following two lines
     - Place an object 50mm in front of the sensor and note the average_reading
     - Calculate the offset value using the formula "offset=50-average_reading"
     - Replace the VL6180X::SYSRANGE__PART_TO_PART_RANGE_OFFSET value of 9 with your offset
       Negative offset values must be expressed in 2's-complement format
     - Restore/uncomment comments when you are happy with the readings.
  */
  //North.writeReg(VL6180X::SYSRANGE__PART_TO_PART_RANGE_OFFSET,0);   // use this offset when calibrating
  North.writeReg(VL6180X::SYSRANGE__PART_TO_PART_RANGE_OFFSET, 9);
  North.stopContinuous();
  delay(300);
  North.startRangeContinuous(50);

  // ------ Configure East_facing sensor
  digitalWrite(East_enable, HIGH);                    // activate sensor
  delay(50);
  East.init();                                        // configure sensor
  East.configureDefault();
  East.setAddress(East_address);                      // change default address
  East.setTimeout(500);
  /*
    ------------
    Calibration:
    ------------
    The VL6180X readings can be up to 15mm out
    To calibrate the sensors:
     - Comment out code lines 153, 162..168, 198..199
     - Uncomment/comment the following two lines
     - Place an object 50mm in front of the sensor and note the average_reading
     - Calculate the offset value using the formula "offset=50-average_reading"
     - Replace the VL6180X::SYSRANGE__PART_TO_PART_RANGE_OFFSET value of 22 with your offset
       Negative offset values must be expressed in 2's-complement format
     - Restore/uncomment comments when you are happy with the readings.
  */
  //East.writeReg(VL6180X::SYSRANGE__PART_TO_PART_RANGE_OFFSET,0);
  East.writeReg(VL6180X::SYSRANGE__PART_TO_PART_RANGE_OFFSET, 22);
  East.stopContinuous();
  delay(300);
  East.startRangeContinuous(50);

  // ----- Connect display
  connect_to_display();
}

// ---------------
//  loop()
// ---------------
void loop() {

  // ----- if data is available to read,
  if (Serial.available() > 0)
  {
    Char = (char)Serial.read();

    // ----- send data whenever a send character ('S') is received
    if (Char == 'S')
    {
      if (!North.timeoutOccurred()) {
        North_reading = (float)North.readRangeContinuousMillimeters() -  North_offset;
        Average0 = moving_average(Array0, sizeof(Array0) / sizeof(Array0[0]), &Sum0, &Index0, North_reading);
      }

      if (!East.timeoutOccurred()) {
        East_reading = (float)East.readRangeContinuousMillimeters() - East_offset;
        Average1 = moving_average(Array1, sizeof(Array1) / sizeof(Array1[0]), &Sum1, &Index1, East_reading);
      }

      calculate_speed_direction(Average0, Average1);

      Average2 = moving_average(Array2, sizeof(Array2) / sizeof(Array2[0]), &Sum2, &Index2, Direction);

      // ----- send results to the display
      Serial.print(Windspeed);
      Serial.print(',');
      Serial.print(Average2);               // Direction
      Serial.print(',');

      /*
        To use these values:
         - Set the North_offset and East_offset in the Arduino header to zero
         - Replace the above offset values with these readings
         - The windspeed indicator should resemble a pin-prick
      */
      Serial.print(Average0 - Radius);      // Required North_offset in still air
      Serial.print(',');
      Serial.println(Average1 - Radius);    // Required East_offset in still air
    }
  }
}

// ----------------------------
// connect to graphics display
// ----------------------------
void connect_to_display()
{
  while (Serial.available() <= 0)
  {
    // ----- keep sending synch ('S') until the display responds
    Serial.println("S");
    delay(100);
  }
}

// -----------------------------
//  calculate_speed_direction()
// -----------------------------
float calculate_speed_direction(float north, float east) {

  // ----- locals
  float N = north;                                      // north reading (adjusted)
  float E = east;                                       // east reading (adjusted)
  float C;                                              // chord formed by laser readings
  float R = Radius;                                     // pipe radius
  float A1;                                             // angle between radius and chord
  float A2;                                             // angle between north and chord
  float A3;                                             // angle between east and chord;
  float A4;                                             // diference angle between north and radius;
  float A5;                                             // difference angle between east and radius;
  float A6;                                             // arc-tangent dY/dX
  float dX;                                             // horizontal offset
  float dY;                                             // vertical offset
  float D;                                              // wind direction

  // ----- calculations
  C = sqrt(sq(N) + sq(E));
  A1 = acos(C / (2 * R)) * RAD_TO_DEG;
  A2 = acos(N / C) * RAD_TO_DEG;
  A3 = 90 - A2;
  A4 = A1 - A2;
  A5 = A1 - A3;
  dY = R * fabs(cos(A4 * DEG_TO_RAD)) - N;
  dX = R * fabs(cos(A5 * DEG_TO_RAD)) - E;
  if (fabs(dX) == 0)dX = 0.0001;
  A6 = fabs(atan(dY / dX) * RAD_TO_DEG);

  // ----- Apply signs to dY and dX
  if (fabs(N - R * cos(A4 * DEG_TO_RAD) - dY) > 0.001) dY = -dY;
  if (fabs(E - R * cos(A5 * DEG_TO_RAD) - dX) > 0.001) dX = -dX;

  // ----- Calculate wind direction
  if ((dX >= 0) && (dY >= 0)) Direction = 270 - A6;       // quadrant 0
  if ((dX < 0) && (dY >= 0)) Direction =  90 + A6;        // quadrant 1
  if ((dX < 0) && (dY < 0)) Direction =   90 - A6;        // quadrant 2
  if ((dX >= 0) && (dY < 0)) Direction =  270 + A6;       // quadrant 3

  // ----- calculate the windspeed
  Windspeed = sqrt(sq(dX) + sq(dY)) * Scale;
}

// ------------------
//  moving_average()
// ------------------
float moving_average(float *array_name, int array_length, float *array_sum, int *array_index, float new_value) {
  *array_sum = *array_sum - array_name[*array_index] + new_value;     // subtract earliest value; add latest value
  array_name[*array_index] = new_value;                               // store latest value in vacant position
  (*array_index)++;                                                   // point to next earliest value
  if (*array_index >= array_length) {
    *array_index = 0;                                                 // wrap around
  }
  return *array_sum / array_length;                                   // moving average
}
